package mainpackage;


import datapackage.*;
import parser.*;
import visitor.*;
import syntaxtree.*;
import java.io.*;
import java.util.HashMap;
import java.util.Map;


public class Main
{

    public static MJClass mainClass;
    public static Map<String, MJClass> programClasses;
    public static String currentFile;
    public static FileWriter writer;

    public static void main(String[] args)
    {
        if (args.length < 1)
        {
            System.err.println("Usage: java Main <file1> <file2> ... <fileN>");
            System.exit(1);
        }

        FileInputStream fis = null;
        for (String arg : args)
        {
            mainClass = null;
            programClasses = new HashMap();
            currentFile = arg;
            writer = null;

            try
            {
                writer = new FileWriter(new File(arg.replace(".java", ".pg")));
                
                fis = new FileInputStream(arg);
                MiniJavaParser parser = new MiniJavaParser(fis);
                Goal tree = parser.Goal();

                ClassVisitor cv = new ClassVisitor();
                MemberVisitor memv = new MemberVisitor();
                MethodVisitor methv = new MethodVisitor();

                tree.accept(cv);
                tree.accept(memv, null);
                tree.accept(methv, null);
                
                //System.out.println("\n////////////File:" + arg + "////////////\n");
                //printSymbolTable();
            }
            catch (ParseException ex)
            {
                System.err.println(ex.getMessage());
            }
            catch (FileNotFoundException ex)
            {
                System.err.println(ex.getMessage());
            }
            catch (Exception ex)
            {
                System.err.println(ex.getMessage());
            }
            finally
            {              
                try
                {
                    if (fis != null)
                    {
                        fis.close();
                    }
                    if (writer != null)
                    {
                        writer.flush();
                        writer.close();
                    }                    
                }
                catch (IOException ex)
                {
                    System.err.println(ex.getMessage());
                }
            }
        }
    }
    
    public static void printSymbolTable()
    {
        System.out.print("Main Class: " + mainClass.getIdentifier()+ "\nFunction: static void main(String[] ");
        for (Map.Entry<String, Variable> argumentsMapEntry : mainClass.getFunctions().get("main").getArguments().entrySet())
        {
            System.out.println(argumentsMapEntry.getValue().getIdentifier() + ")");
        }        
        System.out.println("{");
        for (Map.Entry<String, Variable> variableMapEntry : mainClass.getFunctions().get("main").getVariables().entrySet())
        {
            System.out.println("\tVariable: " + variableMapEntry.getValue().getType() 
                               + " " + variableMapEntry.getValue().getIdentifier());
        }
        System.out.println("}");       
        
        for (Map.Entry<String, MJClass> classMapEntry : programClasses.entrySet())
        {
            System.out.println("\nOther Class: " + classMapEntry.getKey()
                               + (classMapEntry.getValue().getExtendsClass() == null ? "" : " extends " 
                               + classMapEntry.getValue().getExtendsClass()) + "\n");

            for (Map.Entry<String, Variable> variableMapEntry : classMapEntry.getValue().getVariables().entrySet())
            {
                System.out.println("Member" + classMapEntry.getValue().getVariablePosition(variableMapEntry.getValue().getIdentifier())
                                   + "(" + variableMapEntry.getValue().getPigId() + ")"
                                   + ": " + variableMapEntry.getValue().getType() 
                                   + " " + variableMapEntry.getValue().getIdentifier());
            }
            for (Map.Entry<String, Function> functionMapEntry : classMapEntry.getValue().getFunctions().entrySet())
            {
                System.out.print("Function" + classMapEntry.getValue().getFunctionPosition(functionMapEntry.getValue().getIdentifier())
                                 + ": " +  functionMapEntry.getValue().getType() 
                                 + " " + functionMapEntry.getValue().getIdentifier());

                System.out.print("(");
                for (Map.Entry<String, Variable> argumentsMapEntry : functionMapEntry.getValue().getArguments().entrySet())
                {
                    System.out.print(argumentsMapEntry.getValue().getType() + " " 
                                     + argumentsMapEntry.getValue().getIdentifier() + " ("
                                     + argumentsMapEntry.getValue().getPigId() + "), ");
                }
                System.out.println(")");

                System.out.println("{");
                for (Map.Entry<String, Variable> variableMapEntry : functionMapEntry.getValue().getVariables().entrySet())
                {
                    System.out.println("\tVariable" + "(" 
                                       + variableMapEntry.getValue().getPigId()
                                       + ")" + ": " + variableMapEntry.getValue().getType() 
                                       + " " + variableMapEntry.getValue().getIdentifier());
                }
                System.out.println("}");
            }
        }        
    }
}
