package visitor;

import datapackage.*;
import java.util.ArrayList;
import java.util.List;
import mainpackage.*;
import syntaxtree.*;

public class MethodVisitor extends GJDepthFirst<String, Function>
{

    private MJClass currentClass;
    private boolean returnTypeMode = false;
    private boolean returnExp = false;
    private boolean arrayAssign = false;
    private String identifier;
    private int labelCounter;

    @Override
    public String visit(MainClass n, Function argu) throws Exception
    {
        currentClass = Main.mainClass;
          
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
        n.f7.accept(this, argu);
        n.f8.accept(this, argu);
        n.f9.accept(this, argu);
        n.f10.accept(this, argu);
        n.f11.accept(this, argu);
        n.f12.accept(this, argu);
        n.f13.accept(this, argu);

        Main.writer.write("MAIN ");
        n.f14.accept(this, Main.mainClass.getFunctions().get("main"));

        returnTypeMode = true;
        n.f15.accept(this, Main.mainClass.getFunctions().get("main"));
        returnTypeMode = false;

        Main.writer.write("END ");
        
        n.f16.accept(this, argu);
        n.f17.accept(this, argu);

        return null;
    }

    @Override
    public String visit(ClassDeclaration n, Function argu) throws Exception
    {
        currentClass = Main.programClasses.get(n.f1.f0.tokenImage);

        return super.visit(n, argu);
    }

    @Override
    public String visit(ClassExtendsDeclaration n, Function argu) throws Exception
    {
        currentClass = Main.programClasses.get(n.f1.f0.tokenImage);

        return super.visit(n, argu);
    }

    @Override
    public String visit(VarDeclaration n, Function argu) throws Exception
    {
        String type = n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);

        if (argu != null)
        {
            argu.getVariables().put(n.f1.f0.tokenImage, new Variable(n.f1.f0.tokenImage, type, false));                        
        }

        return null;
    }

    @Override
    public String visit(MethodDeclaration n, Function argu) throws Exception
    {               
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        
        Function curFun = currentClass.getFunctions().get(n.f2.f0.tokenImage);
        Variable.idCounter = curFun.getArguments().size() + 1;
        labelCounter = 0;
        
        Main.writer.write(currentClass.getIdentifier() + "_" + curFun.getIdentifier()
                          + " [ " + (curFun.getArguments().size() + 1) + " ] BEGIN ");
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
                             
        n.f7.accept(this, curFun);        
        
        returnTypeMode = true;
        n.f8.accept(this, curFun);
        returnTypeMode = false;
        
        n.f9.accept(this, argu);
        
        Main.writer.write("RETURN ");
        returnExp = true;
        n.f10.accept(this, curFun);
        returnExp = false;
        Main.writer.write("END ");
        
        n.f11.accept(this, argu);
        n.f12.accept(this, argu);

        return null;
    }    

    @Override
    public String visit(ArrayType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "array";
    }

    @Override
    public String visit(BooleanType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "boolean";
    }

    @Override
    public String visit(IntegerType n, Function argu) throws Exception
    {
        super.visit(n, argu);

        return "int";
    }

    @Override
    public String visit(AssignmentStatement n, Function argu) throws Exception
    {                        
        Variable var = getVar(n.f0.f0.tokenImage, argu);
        // If the variable for the assignment is a field of the class        
        if(var.getPigId().equals(""))
        {
            Main.writer.write("HSTORE ");            
        }
        else
        {
            Main.writer.write("MOVE ");            
        }        
        
        returnTypeMode = false;
        n.f0.accept(this, argu);
        returnTypeMode = true;                       
       
        n.f1.accept(this, argu);  
        
        returnExp = true;
        n.f2.accept(this, argu); 
        returnExp = false;
        
        n.f3.accept(this, argu);        
        
        return null;
    }  

    @Override
    public String visit(ArrayAssignmentStatement n, Function argu) throws Exception
    {       
        returnExp = true;
        arrayAssign = true;
        n.f0.accept(this, argu);               
        returnExp = false;
        arrayAssign = false;
        String curId = identifier;
             
        // If it's a class member array
        if(curId.startsWith("BEGIN"))
        {
            int arrId = Variable.idCounter;
            Variable.idCounter++;
            Main.writer.write("HSTORE PLUS BEGIN HLOAD TEMP " + arrId + " " + curId + " 0 " + 
                              "RETURN TEMP " + arrId + " END ");
        }
        else
        {
            Main.writer.write("HSTORE PLUS " + curId + " ");
        }
        
        
        n.f1.accept(this, argu);              
        
        int indexId = Variable.idCounter;
        Variable.idCounter++;                
        Main.writer.write("PLUS BEGIN MOVE TEMP " + indexId + " TIMES ");        
        
        returnExp = true;
        n.f2.accept(this, argu);  
        returnExp = false;
        
        int errId = Variable.idCounter;
        Variable.idCounter++;                  
        Main.writer.write("4 HLOAD TEMP " + errId + " ");

        if(curId.startsWith("BEGIN"))
        {
            int arrId = Variable.idCounter;
            Variable.idCounter++;            
            Main.writer.write("BEGIN HLOAD TEMP " + arrId + " " + curId + " 0 " + 
                              "RETURN TEMP " + arrId + " END ");
        }
        else
        {
            Main.writer.write(curId + " ");
        }        
        
        int elseLabel1 = labelCounter;
        labelCounter++;
        int elseLabel2 = labelCounter;
        labelCounter++;        
        Main.writer.write(" 0 " + 
                          "CJUMP MINUS 1 LT " + identifier + " TEMP " + errId + " L" + elseLabel1 + 
                          " ERROR L" + elseLabel1 + " NOOP " +
                          "CJUMP LT " + identifier + " 0 L" + elseLabel2 + 
                          " ERROR L" + elseLabel2 + " NOOP " +
                          "RETURN TEMP " + indexId + " END 4 0 ");
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
        
        return null;
    }

    @Override
    public String visit(IfStatement n, Function argu) throws Exception
    {
        Main.writer.write("CJUMP ");
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        
        returnExp = true;
        n.f2.accept(this, argu);
        returnExp = false;

        int elseLabel = labelCounter;
        labelCounter++;
        Main.writer.write("L" + elseLabel + " ");        
        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        
        int endLabel = labelCounter;
        labelCounter++;                
        Main.writer.write("JUMP L" + endLabel + " L" + elseLabel + " ");
        
        n.f5.accept(this, argu);
        n.f6.accept(this, argu);
        
        Main.writer.write("L" + endLabel + " NOOP ");
        
        return null;
    }

    @Override
    public String visit(WhileStatement n, Function argu) throws Exception
    {
        int whileLabel = labelCounter;
        labelCounter++;
        Main.writer.write("L" + whileLabel + " CJUMP ");
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);  
        
        returnExp = true;
        n.f2.accept(this, argu);        
        returnExp = false;
        
        n.f3.accept(this, argu);
        
        int endLabel = labelCounter;
        labelCounter++;
        Main.writer.write("L" + endLabel + " ");        
        
        n.f4.accept(this, argu);
        
        Main.writer.write("JUMP L" + whileLabel + " L" + endLabel + " NOOP ");        
        
        return null;
    }

    @Override
    public String visit(PrintStatement n, Function argu) throws Exception
    {
        Main.writer.write("PRINT ");
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);        
        n.f2.accept(this, argu);        
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        
        return null;
    }    

    @Override
    public String visit(AndExpression n, Function argu) throws Exception
    {
        int varId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("BEGIN MOVE TEMP " + varId + " 0 CJUMP ");       
                
        n.f0.accept(this, argu);       
        
        int endLabel = labelCounter;
        labelCounter++;        
        Main.writer.write("L" + endLabel + " CJUMP ");
        
        n.f1.accept(this, argu);        
        n.f2.accept(this, argu);        
        
        Main.writer.write("L" + endLabel + " MOVE TEMP " + varId + " 1 "
                          + "L" + endLabel + " NOOP RETURN TEMP " + varId + " END ");

        return "boolean";
    }

    @Override
    public String visit(CompareExpression n, Function argu) throws Exception
    {
        Main.writer.write("LT ");
        
        n.f0.accept(this, argu);        
        n.f1.accept(this, argu);        
        n.f2.accept(this, argu);
        
        return "boolean";
    }

    @Override
    public String visit(PlusExpression n, Function argu) throws Exception
    {
        int varId = -1;
        if(returnExp)
        {
            varId = Variable.idCounter;
            Variable.idCounter++;
            Main.writer.write("BEGIN MOVE TEMP " + varId + " ");
        }
        Main.writer.write("PLUS ");
        
        n.f0.accept(this, argu);        
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        
        if(returnExp)
        {
            Main.writer.write("RETURN TEMP " + varId + " END ");
            identifier = "TEMP " + varId;
        }        

        return "int";        
    }

    @Override
    public String visit(MinusExpression n, Function argu) throws Exception
    {
        int varId = -1;
        if(returnExp)
        {
            varId = Variable.idCounter;
            Variable.idCounter++;
            Main.writer.write("BEGIN MOVE TEMP " + varId + " ");
        }               
        Main.writer.write("MINUS ");
        
        n.f0.accept(this, argu);        
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        
        if(returnExp)
        {
            Main.writer.write("RETURN TEMP " + varId + " END ");
            identifier = "TEMP " + varId;
        }         
        
        return "int";
    }

    @Override
    public String visit(TimesExpression n, Function argu) throws Exception
    {
        int varId = -1;
        if(returnExp)
        {
            varId = Variable.idCounter;
            Variable.idCounter++;
            Main.writer.write("BEGIN MOVE TEMP " + varId + " ");
        }        
        Main.writer.write("TIMES ");
        
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        
        if(returnExp)
        {
            Main.writer.write("RETURN TEMP " + varId + " END ");
            identifier = "TEMP " + varId;
        }         
        
        return "int";
    }

    @Override
    public String visit(ArrayLookup n, Function argu) throws Exception
    {
        int varId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("BEGIN HLOAD TEMP " + varId + " PLUS ");
        
        returnExp = true;
        n.f0.accept(this, argu);      
        returnExp = false;
        String curId = identifier;
        
        n.f1.accept(this, argu);        

        int indexId = Variable.idCounter;
        Variable.idCounter++;        
        Main.writer.write("PLUS BEGIN MOVE TEMP " + indexId + " TIMES ");        
        
        returnExp = true;
        n.f2.accept(this, argu);                
        returnExp = false;
        
        n.f3.accept(this, argu);

        int errId = Variable.idCounter;
        Variable.idCounter++;
        int elseLabel1 = labelCounter;
        labelCounter++;
        int elseLabel2 = labelCounter;
        labelCounter++;        
        Main.writer.write("4 HLOAD TEMP " + errId + " " + curId + " 0 " +
                          " CJUMP MINUS 1 LT " + identifier + " TEMP " + errId +
                          " L" + elseLabel1 + " ERROR L" + elseLabel1 + " NOOP " +
                          "CJUMP LT " + identifier + " 0 "+
                          "L" + elseLabel2 + " ERROR L" + elseLabel2 + " NOOP " +                          
                          "RETURN TEMP " + indexId + " END 4 0 "+
                          "RETURN TEMP " + varId + " END ");

        return "int";
    }

    @Override
    public String visit(ArrayLength n, Function argu) throws Exception
    {               
        int lenId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("BEGIN HLOAD TEMP " + lenId + " " );
        
        returnExp = true;
        n.f0.accept(this, argu);               
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        returnExp = false;
        
        Main.writer.write(" 0 RETURN TEMP " + lenId + " END ");

        return "int";
    }

    @Override
    public String visit(MessageSend n, Function argu) throws Exception
    {
        int objId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("CALL BEGIN MOVE TEMP " + objId + " ");
        
        String type = n.f0.accept(this, argu);        
        n.f1.accept(this, argu);
        
        returnTypeMode = false;
        n.f2.accept(this, argu);
        returnTypeMode = true;                  

        int vtId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("HLOAD TEMP " + vtId + " TEMP " + objId + " 0 ");
        int funId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("HLOAD TEMP " + funId + " TEMP " + vtId +
                          " " + Main.programClasses.get(type).getFunctionPosition(n.f2.f0.tokenImage) + " ");
        Main.writer.write("RETURN TEMP " + funId + " END (TEMP " + objId + " ");
        
        n.f3.accept(this, argu);       
        n.f4.accept(this, argu);                  
        n.f5.accept(this, argu);
        
        Main.writer.write(") ");
        
        return Main.programClasses.get(type).getFunctions().get(n.f2.f0.tokenImage).getType();
    }    
    
    @Override
    public String visit(IntegerLiteral n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        Main.writer.write(n.f0.tokenImage + " ");
        identifier = n.f0.tokenImage;

        return "int";
    }

    @Override
    public String visit(TrueLiteral n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        Main.writer.write("1 ");        

        return "boolean";
    }

    @Override
    public String visit(FalseLiteral n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        Main.writer.write("0 ");

        return "boolean";
    }

    @Override
    public String visit(Identifier n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        if(argu != null)
        {
            Variable var = getVar(n.f0.tokenImage, argu);
            if(var != null)
            {
                if(var.getPigId().equals(""))
                {
                    if(returnExp)
                    {
                        if(var.getType().equals("array") && arrayAssign)
                        {
                            int varId = Variable.idCounter;
                            Variable.idCounter++;
                            identifier = "BEGIN MOVE TEMP " + varId + 
                                         " PLUS TEMP 0 " + currentClass.getVariablePosition(var.getIdentifier()) +
                                         " RETURN TEMP " + varId + " END ";                            
                        }
                        else
                        {
                            int varId = Variable.idCounter;
                            Variable.idCounter++;
                            identifier = "BEGIN HLOAD TEMP " + varId + 
                                         " TEMP 0 " + currentClass.getVariablePosition(var.getIdentifier()) +
                                         " RETURN TEMP " + varId + " END ";
                            Main.writer.write(identifier);
                        }                                                
                    }
                    else
                    {
                        identifier = "TEMP 0 " + currentClass.getVariablePosition(var.getIdentifier()) + " ";
                        Main.writer.write(identifier);                                    
                    }
                }
                else
                {
                    identifier = var.getPigId() + " ";
                    if(!arrayAssign)
                    {                        
                        Main.writer.write(identifier);
                    }
                }        
            }
        }
        
        if(returnTypeMode)
        {            
            String type = getVar(n.f0.tokenImage, argu).getType();            
            return type;
        }

        return n.f0.tokenImage;
    }

    @Override
    public String visit(ThisExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        Main.writer.write("TEMP 0 ");

        return currentClass.getIdentifier();
    }

    @Override
    public String visit(ArrayAllocationExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);              
        
        int arrId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("BEGIN MOVE TEMP " + arrId + " HALLOCATE TIMES PLUS ");                
        
        returnExp = true;
        n.f3.accept(this, argu);   
        returnExp = false;
        
        int countId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write(" 1 4 " + "MOVE TEMP " + countId + " 4 ");        
        int whileLabel = labelCounter;
        labelCounter++;
        int endLabel = labelCounter;
        labelCounter++;        
        int elseLabel = labelCounter;
        labelCounter++;                
        Main.writer.write("CJUMP LT " + identifier + " 0 "+
                          "L" + elseLabel + " ERROR L" + elseLabel + " NOOP " +
                          "L" + whileLabel + " CJUMP LT TEMP " + countId +
                          " TIMES PLUS " + identifier + " 1 4 L" + endLabel + " " + 
                          "HSTORE PLUS TEMP " + arrId + " TEMP " + countId + " 0 0 " + 
                          "MOVE TEMP " + countId + " PLUS TEMP " + countId + " 4 " + 
                          "JUMP L" + whileLabel + " L" + endLabel + " " +
                          "HSTORE TEMP " + arrId + " 0 " + identifier + " " +
                          "RETURN TEMP " + arrId + " END ");
        
        n.f4.accept(this, argu);
        
        return "array";
    }

    @Override
    public String visit(AllocationExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);        
        
        returnTypeMode = false; 
        n.f1.accept(this, argu);
        returnTypeMode = true;        
                   
        MJClass newClass = Main.programClasses.get(n.f1.f0.tokenImage);        
        List<MJClass> extendsList = newClass.getExtendsList();                            

        List<Function> funList = new ArrayList();
        List<Function> curList = new ArrayList();
        List<String> funStore = new ArrayList();
        
        int vtId = Variable.idCounter;
        Variable.idCounter++;  
        for(int i = (extendsList.size() - 1); i >= 0; i--)
        {
            for(int j = 0; j < extendsList.get(i).getFunctions().size(); j++)
            {             
                curList.clear();
                curList.addAll(extendsList.get(i).getFunctions().values());
                for(int c = 0; c < curList.size(); c++)
                {
                    int f;
                    for(f = 0; f < funList.size(); f++)
                    {
                        if(curList.get(c).getIdentifier().equals(funList.get(f).getIdentifier()))
                        {
                            funStore.set(f, "HSTORE TEMP " + vtId + " " + (f * 4) + " " +
                                         extendsList.get(i).getIdentifier() + "_" + curList.get(c).getIdentifier() + " ");                            
                            break;
                        }
                    }
                    
                    if(f == funList.size())
                    {
                        funList.add(curList.get(c));
                        funStore.add("HSTORE TEMP " + vtId + " " + ((funList.size() - 1) * 4) + " " +
                                     extendsList.get(i).getIdentifier() + "_" + curList.get(c).getIdentifier() + " ");                        
                    }
                }
            }
        }                
        Main.writer.write("BEGIN MOVE TEMP " + vtId + " HALLOCATE " + (funList.size() * 4)); 


        List<Variable> varList = new ArrayList();             
        for(int i = (extendsList.size() - 1); i >= 0; i--)
        {
            varList.addAll(extendsList.get(i).getVariables().values());            
        }         
        
        int objId = Variable.idCounter;
        Variable.idCounter++;                  
        Main.writer.write(" MOVE TEMP " + objId + " HALLOCATE " + 
                          ((varList.size() + 1) * 4) + " ");
                
        for(int i = 0; i < funStore.size(); i++)
        {
            Main.writer.write(funStore.get(i));
        }
        
        if(!varList.isEmpty())
        {
            int countId = Variable.idCounter;
            Variable.idCounter++;
            Main.writer.write("MOVE TEMP " + countId + " 4 "); 
            int whileLabel = labelCounter;
            labelCounter++;
            int endLabel = labelCounter;
            labelCounter++;
            Main.writer.write("L" + whileLabel + " CJUMP LT TEMP " + countId + " " +
                              ((varList.size() + 1) * 4) + " L" + endLabel + 
                              " HSTORE PLUS TEMP " + objId + " TEMP " + countId + " 0 0 " + 
                              "MOVE TEMP " + countId + " PLUS TEMP " + countId + " 4 " + 
                              "JUMP L" + whileLabel +
                              " L" + endLabel + " ");
        }
        Main.writer.write("HSTORE TEMP " + objId + " 0 TEMP " + vtId +
                          " RETURN TEMP " + objId + " END ");
        
        n.f2.accept(this, argu);                
        n.f3.accept(this, argu);                        
        
        return n.f1.f0.tokenImage;
    }

    @Override
    public String visit(NotExpression n, Function argu) throws Exception
    {
        Main.writer.write("MINUS 1 ");
        
        n.f0.accept(this, argu);        
        n.f1.accept(this, argu);                
        
        return "boolean";
    }

    @Override
    public String visit(BracketExpression n, Function argu) throws Exception
    {
        n.f0.accept(this, argu);
        
        int expId = Variable.idCounter;
        Variable.idCounter++;
        Main.writer.write("BEGIN MOVE TEMP " + expId + " ");
        
        String type = n.f1.accept(this, argu);
                
        Main.writer.write("RETURN TEMP " + expId + " END ");
        identifier = "TEMP " + expId;     
        
        n.f2.accept(this, argu);
        
        return type;
    }   

    public Variable getVar(String identifier, Function argu) throws Exception
    {        
        Variable var = null;

        if (argu.getVariables().containsKey(identifier))
        {
            var = argu.getVariables().get(identifier);
        }
        else if (argu.getArguments().containsKey(identifier))
        {
            var = argu.getArguments().get(identifier);
        }
        else if (currentClass.getVariables().containsKey(identifier))
        {
            var = currentClass.getVariables().get(identifier);
        }
        else
        {
            MJClass tempClass = currentClass;
            
            while(tempClass.getExtendsClass() != null 
                  && !tempClass.getExtendsClass().equals(Main.mainClass.getIdentifier()))
            {
                if(Main.programClasses.get(tempClass.getExtendsClass()).getVariables().containsKey(identifier))
                {                
                    var = Main.programClasses.get(tempClass.getExtendsClass()).getVariables().get(identifier);
                    break;
                }
                else
                {
                    tempClass = Main.programClasses.get(tempClass.getExtendsClass());
                }
            }            
        }                
        
        return var;
    } 
}
